"""Top-level package for Split Dataset."""

__author__ = """Vilim Stih & Luigi Petrucco @portugueslab"""
__version__ = '0.2.2'

from split_dataset.blocks import Blocks
from split_dataset.split_dataset import SplitDataset, EmptySplitDataset, save_to_split_dataset



import numpy as np
from pathlib import Path
import json
import flammkuchen as fl
from split_dataset.blocks import Blocks
from itertools import product

# TODO this should probably be done as a constructor of the SplitDataset
def save_to_split_dataset(
    data,
    root_name,
    block_size=None,
    crop=None,
    padding=0,
    prefix="",
    compression="blosc",
):
    """ Function to save block of data into a split_dataset.
    """

    new_name = prefix + ("_cropped" if crop is not None else "")
    padding = (
        data.padding if padding is not None and isinstance(data, Blocks) else padding
    )
    blocks = EmptySplitDataset(
        shape_full=data.shape,
        shape_block=data.shape_block if block_size is None else block_size,
        crop=crop,
        padding=padding,
        root=root_name,
        name=new_name,
    )
    for filename, (idxs, slices) in zip(blocks.files, blocks.slices()):
        fl.save(
            str(blocks.root / filename),
            {"stack_{}D".format(len(blocks.shape_cropped)): data[slices]},
            compression=compression,
        )

    return blocks.finalize()


class SplitDataset(Blocks):
    """
    Manages datasets split over multiple h5 file across arbitrary dimensions.
    To do so, uses the BlockSplitter class functions, and define blocks as
    files.

    """

    def __init__(self, root, prefix=None):
        """

        :param files: A list of files within the root directory containing
        the stack
        :param shape_full: full size of the stack
        :param shape_block: size of a single file
        :param provenance: provenance data
        :param root: The directory containing the files
        """

        # Load information about stack and splitting. Use the json metadata
        # file if possible:
        self.root = Path(root)
        try:
            stack_meta_f = next(self.root.glob("*stack_metadata.json"))

            with open(str(stack_meta_f), "r") as f:
                block_metadata = json.load(f)
        except StopIteration:
            last_data_f = sorted(list(self.root.glob("{}*.h5".format(prefix))))[-1]
            block_metadata = fl.load(str(last_data_f), "/stack_metadata")

            # Ugly keyword fix to handle transition to new json system:
            for new_k, old_k in zip(
                ["shape_block", "shape_full"], ["block_size", "full_size"]
            ):
                block_metadata[new_k] = block_metadata.pop(old_k)

            # By putting this here, we generate the proper stack_metadata
            # file when we open old version data (int conversion for some
            # weird format problem with flammkuchen dictionary):
            clean_metadata = dict()
            _save_metadata_json(block_metadata, self.root)
            for k in block_metadata.keys():
                if isinstance(block_metadata[k], tuple):
                    clean_metadata[k] = tuple(
                        int(n) if n is not None else None for n in block_metadata[k]
                    )
                else:
                    clean_metadata[k] = block_metadata[k]
            with open(str(), "w") as f:
                json.dump(clean_metadata, f)

        # Start the parent BlockSplitter:
        super().__init__(
            shape_full=block_metadata["shape_full"],
            shape_block=block_metadata["shape_block"],
        )
        if prefix != None:
            files = sorted(self.root.glob("*{}_[0-9]*.h5".format(prefix)))
        else:
            files = sorted(self.root.glob("*[0-9]*.h5"))
        self.files = np.array(files).reshape(self.block_starts.shape[:-1])

        # To migrate smoothly to removal of stack_ND key in favour of only stack:
        try:
            self.data_key = [k for k in fl.meta(files[0]).keys() if "stack" in k][0]
        except IndexError:
            self.data_key = None

        # If available, read resolution
        try:
            self.resolution = block_metadata["resolution"]
        except KeyError:
            self.resolution = (1, 1, 1)
        # TODO check this
        self.shape = self.shape_cropped

    def __getitem__(self, item):
        """
        Implement usage of the H5SplitDataset as normal numpy array.
        :param item:
        :return:
        """
        # Lot of input munging to emulate indexing in numpy array
        if np.any(self.padding) != 0:
            raise ValueError(
                "Indexing in datasets with overlap (padding) is"
                " not supported, merge them first with an"
                " appropriate merging function"
            )

        if isinstance(item, int):
            item = (slice(item, item + 1),)

        if isinstance(item, slice):
            item = (item,)

        if isinstance(item, tuple):
            # Take care of the case when only the first few dimensions
            # are specified:
            if len(item) < len(self.shape):
                item = item + (None,) * (len(self.shape) - len(item))

            # Loop over dimensions creating a list of starting and ending
            # points

            starts = []
            ends = []
            singletons = np.zeros(len(item), dtype=np.bool)
            for i_dim, (dim_slc, dim_full) in enumerate(zip(item, self.shape)):
                # i_dim: index of current dimension
                # dim_slc: slice/index for current dimension
                # fd: length of dataset on current dimension

                # If nothing specified, start from 0 and finish at end:
                if dim_slc is None:
                    starts.append(0)
                    ends.append(dim_full)

                # If a slice is specified:
                elif isinstance(dim_slc, slice):
                    if dim_slc.start is None:
                        starts.append(0)
                    else:
                        if dim_slc.start >= 0:
                            starts.append(dim_slc.start)
                        else:
                            starts.append(max(0, dim_full + dim_slc.start))

                    if dim_slc.stop is None:
                        ends.append(dim_full)
                    else:
                        if dim_slc.stop >= 0:
                            ends.append(min(dim_slc.stop, dim_full))
                        else:
                            ends.append(max(0, dim_full + dim_slc.stop))
                elif isinstance(dim_slc, int) or isinstance(dim_slc, np.int32):
                    singletons[i_dim] = True
                    if dim_slc >= 0:
                        if dim_slc > dim_full - 1:
                            raise IndexError(
                                "Indexes {} out of dimensions {}!".format(
                                    item, self.shape
                                )
                            )
                        starts.append(dim_slc)
                        ends.append(dim_slc + 1)
                    else:
                        if -dim_slc > dim_full:
                            raise IndexError(
                                "Indexes {} out of dimensions {}!".format(
                                    item, self.shape
                                )
                            )
                        starts.append(dim_full + dim_slc)
                        ends.append(dim_full + dim_slc + 1)
                else:
                    raise IndexError("Unsupported indexing")
        else:
            raise IndexError("Unsupported indexing")

        file_slices, take_block_s_idx, take_block_e_idx = self.blocks_to_take(
            starts, ends
        )
        output_size = tuple(e - s for s, e in zip(starts, ends))

        output = None

        # A lot of indexing tricks to achieve multidimensional generality
        for f_idx in product(*(range(s, e) for s, e in file_slices)):
            abs_idx = [ri - s for ri, (s, e) in zip(f_idx, file_slices)]
            sel_slices = tuple(
                slice(0 if ci != s else si, None if ci < e - 1 else ei)
                for ci, (s, e), si, ei in zip(
                    f_idx, file_slices, take_block_s_idx, take_block_e_idx
                )
            )
            arr = fl.load(
                str(self.files[f_idx]), "/" + self.data_key, sel=fl.aslice[sel_slices],
            )

            if output is None:
                output = np.empty(output_size, arr.dtype)

            output_sel_tuple = tuple(
                slice(
                    0 if st_idx == 0 else bs - first_idx + (st_idx - 1) * (bs),
                    (0 if st_idx == 0 else bs - first_idx + (st_idx - 1) * (bs)) + sz,
                )
                for st_idx, bs, first_idx, sz in zip(
                    abs_idx, self.shape_block, take_block_s_idx, arr.shape
                )
            )
            output[output_sel_tuple] = arr

        if output is None:
            raise IndexError(
                "Trying to index the split dataset outside of bounds, between "
                + str(starts)
                + " and "
                + str(ends)
            )

        output_sel = tuple(0 if singleton else slice(None) for singleton in singletons)

        return output[output_sel]

    def apply_crop(self, crop):
        """ Take out the data with a crop

        """
        # TODO there is the crop atrribute, which is a lazy crop, this should actually return a non-cropped dataset
        ds_cropped = EmptySplitDataset(
            shape_full=self.shape,
            shape_block=self.shape_block,
            padding=self.padding,
            crop=crop,
            root=self.root.parent,
            name=self.root.name + "_cropped",
        )
        # the slices iterator does not return just the slices, but also the indicesS
        for (i_slice, block_slices), file_name in zip(
            ds_cropped.slices(), ds_cropped.files
        ):
            fl.save(
                str(self.root / file_name),
                {"stack_{}D".format(ds_cropped.n_dims): self[block_slices]},
            )

        ds_cropped.finalize()


class EmptySplitDataset(Blocks):
    """ Class to initialize an empty dataset for which we have to save metadata
    after filling its blocks.
    """

    def __init__(self, root, name, *args, resolution=None, **kwargs):
        """
        :param root: folder where the stack will be saved;
        :param name: name of the dataset, for the folder name;
        :param resolution: resolution of the stack, in microns;
        """
        super().__init__(*args, **kwargs)
        self.root = Path(root) / name
        if not self.root.is_dir():
            self.root.mkdir(parents=True)
        self.files = ["{:04d}.h5".format(i) for i in range(self.n_blocks)]
        self.resolution = resolution

    def save_block_data(self, n, data, verbose=False):
        """ Optional method to save data in a block. Often we don't use it,
        as we directly save data in the parallelized function. Might be good to
        find ways of centralizing saving here?
        :param n: n of the block we are saving in;
        :param data: data to be pured in the block;
        :param verbose:
        :return:
        """
        fname = "{:04d}.h5".format(n)
        if verbose:
            print("Saving ", str(self.root / fname))

            if data.shape != self.shape_block:
                print(" - data has different dimension from block!")

        to_save = {f"stack": data}

        fl.save(str(self.root / fname), to_save, compression="blosc")

    def finalize(self):
        n_dims = len(self.shape_block)
        block_dict = self.serialize()
        block_dict["shape_full"] = self.shape_cropped
        block_dict["crop_start"] = (0,) * n_dims
        block_dict["crop_end"] = (0,) * n_dims
        block_dict["resolution"] = self.resolution \
            if self.resolution is not None else (1,) * n_dims

        block_dict["axis_order"] = "tzyx" if n_dims == 4 else "zyx"

        _save_metadata_json(block_dict, self.root)
        return SplitDataset(self.root)


def _save_metadata_json(dictionary, root):
    """ Save json file preventing type failures for stack shapes
    :param path: path for saving
    :param dictionary: dictionary to be saved
    :return:
    """
    METADATA_FILENAME = "stack_metadata.json"
    for k in dictionary.keys():
        if type(dictionary[k]) is tuple:
            # funny fix for variable type mysterious error:
            if type(dictionary[k][0]) == np.int64 or type(dictionary[k][0]) == int:
                dictionary[k] = tuple([int(i) for i in dictionary[k]])

    json.dump(dictionary, open(root / METADATA_FILENAME, "w"))



import numpy as np
from split_dataset import Blocks


def test_cartesian_blocks():
    test_size = (20, 20)
    a = np.ones(test_size)
    blocks = Blocks(test_size, shape_block=(3, 7), padding=(1, 2))
    for idx, block in blocks.slices():
        a[block] = 0
    np.testing.assert_array_equal(a, np.zeros(test_size))


def test_dropped_dimension():
    test_size = (5, 15, 20)
    blocks = Blocks(
        test_size, shape_block=(3, 7), padding=(1, 2), crop=((1, 1), (0, 0), (0, 0))
    )
    np.testing.assert_equal(blocks.drop_dim(1).shape_full, (5, 20))



import numpy as np
import flammkuchen as fl
from itertools import product
from typing import Union, Tuple, Optional


class BlockIterator:
    def __init__(self, blocks, slices=True):
        self.blocks = blocks
        self.current_block = 0
        self.slices = slices

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_block == self.blocks.n_blocks:
            raise StopIteration
        else:
            idx = self.blocks.linear_to_cartesian(self.current_block)
            self.current_block += 1
            if self.slices:
                return (
                    idx,
                    tuple(
                        slice(s, e)
                        for s, e in zip(
                            self.blocks.block_starts[idx], self.blocks.block_ends[idx]
                        )
                    ),
                )
            else:
                return (
                    idx,
                    tuple(
                        (s, e)
                        for s, e in zip(
                            self.blocks.block_starts[idx], self.blocks.block_ends[idx]
                        )
                    ),
                )


def _make_iterable(input_var, n_rep=1):
    try:
        iter(input_var)
        return input_var
    except TypeError:
        return (input_var,) * n_rep


class Blocks:
    """
    Blocks have two indexing systems:
     - linear:
     - cartesian: gives the position of the block in the general block tiling.
    """

    def __init__(
        self,
        shape_full: Tuple,
        shape_block: Optional[Tuple] = None,
        dim_split: Optional[int] = None,
        blocks_number: Optional[int] = None,
        padding: Union[int, Tuple] = 0,
        crop: Optional[Tuple] = None,
    ):
        """ Make a block structure. It can be defined using block size or number
        of blocks (number of blocks if specified will overwrite size).
        For example, one split over the 2nd and 3rd dimensions of a 100x20x40x10 block can
        equivalently defined as:
        BlockSplitter((100,20,40,10), block_size=(10,10,20,30))
        BlockSplitter((100,20,40,10), blocks_number=(1, 2, 2, 1))
        BlockSplitter((100,20,40,10), dim_split=(1,2), block_size=(10,20))
        BlockSplitter((100,20,40,10), dim_split=(1,2), blocks_number=(2,2))

        :param shape_full: dimensions of the whole stack
        :param dim_split: dimension along which to split (if undefined, start
           counting from the first dimension)
        :param shape_block: size of blocks along each dimension
        :param blocks_number: number of blocks along each dimension
        :param padding: amount of overlap between blocks
        :param crop: iterable of tuples giving the amount of cropping in
            each dimension
        """
        self._shape_full = shape_full

        if crop is None:
            crop = ((0, 0),) * len(shape_full) if shape_full is not None else None

        self._crop = crop
        self.shape_cropped = shape_full

        self.starts = None
        self.block_starts = None
        self.block_ends = None

        self.update_stack_dims()

        # Define shape block and padding allowing multiple input types.

        # Initialize block size as full stack size and 0 padding:
        self._shape_block = list(self.shape_cropped)
        self._padding = [0 for _ in range(len(self.shape_cropped))]

        if not dim_split:
            dim_split = [j for j, d in enumerate(shape_block) if d is not None]

        # Make tuple if single numbers
        self.dim_split = _make_iterable(dim_split)
        shape_block = _make_iterable(shape_block, max(self.dim_split) + 1)
        pad_amount = _make_iterable(padding, max(self.dim_split) + 1)

        if blocks_number:  # define from required number of blocks
            shape_block = []
            blocks_number = _make_iterable(blocks_number, len(self.dim_split))
            for dim, n in zip(self.dim_split, blocks_number):
                shape_block.append(int(np.ceil(self.shape_cropped[dim] / n)))

        for dim in self.dim_split:
            self._shape_block[dim] = min(shape_block[dim], self.shape_cropped[dim])
            self._padding[dim] = pad_amount[dim]

        self.n_blocks = 0
        self.n_dims = len(shape_full)
        self.shape_block = tuple(self._shape_block)

    @property
    def shape_full(self):
        return self._shape_full

    @shape_full.setter
    def shape_full(self, value):
        self._shape_full = value
        self.update_stack_dims()
        self.update_block_structure()

    @property
    def crop(self):
        return self._crop

    @crop.setter
    def crop(self, value):
        if value is None:
            value = ((0, 0),) * len(self.shape_full)
        self._crop = value
        self.update_stack_dims()
        self.update_block_structure()

    @property
    def shape_block(self):
        return self._shape_block

    @shape_block.setter
    def shape_block(self, value):
        self._shape_block = value
        self.update_block_structure()

    @property
    def padding(self):
        return self._padding

    @padding.setter
    def padding(self, value):
        self._padding = value
        self.update_block_structure()

    def update_stack_dims(self):
        """ Update stack dimensions and cropping, if shape_full or cropping
        is changed.
        :return:
        """

        if self.shape_full is not None:
            self.shape_cropped = tuple(
                d - cl - ch for d, (cl, ch) in zip(self.shape_full, self.crop)
            )
            self.starts = tuple(cl for cl, ch in self.crop)

    def update_block_structure(self):
        """
        Update the Blocks structure, e.g. when block
        shape or padding are changed.
        """
        # Cartesian product for generating a list of indexes on every split
        # dimension (i.e., dimensions where int(np.ceil(stack_size / block_size)
        #  is != 1).
        # For example, splitting one time in 2nd and 3rd dims,
        # idx_blocks = (0, 0, 0, 0), (0, 0, 1, 0), (0, 1, 0, 0), (0, 1, 1, 0).

        # block_starts and block_ends will be arrays of shape
        # (n_blocks_dim0, n_blocks_dim1, n_blocks_dim2 ..., shape_full)
        # by addressing the N-1 dimensions with the index of the block we
        # will get a vector with the starting position of the block on all
        # original dimensions of the full stack.
        if self.shape_block is not None:
            self.block_starts = np.empty(
                tuple(
                    int(np.ceil((stack_size - pad_size) / block_size))
                    for stack_size, block_size, pad_size in zip(
                        self.shape_cropped, self.shape_block, self.padding
                    )
                )
                + (len(self.shape_cropped),),
                dtype=np.int32,
            )
            self.block_ends = np.empty_like(self.block_starts)
            for idx_blocks in product(*(range(s) for s in self.block_starts.shape[:-1])):
                self.block_starts[idx_blocks + (slice(None),)] = [
                    st + i_bd * bs
                    for i_bd, bs, st in zip(idx_blocks, self.shape_block, self.starts)
                ]
                self.block_ends[idx_blocks + (slice(None),)] = [
                    min(maxdim + st, (i_bd + 1) * bs + pd + st)
                    for i_bd, bs, pd, maxdim, st in zip(
                        idx_blocks,
                        self.shape_block,
                        self.padding,
                        self.shape_cropped,
                        self.starts,
                    )
                ]

            # for iterations
            self.n_blocks = np.product(self.block_starts.shape[:-1])
            self.n_dims = len(self.shape_cropped)

    def slices(self, as_tuples=False):
        return BlockIterator(self, slices=not as_tuples)

    def linear_to_cartesian(self, lin_idx):
        """
        Convert block linear index into cartesian index.
        Example: in a 3D stack split in 2x2x3 blocks,

        self.linear_to_cartesian(0) = (0,0,0)  # first block
        bs.linear_to_cartesian(11) = (1,1,2)  # last block
        :param lin_idx: block linear index (int)
        :return: block cartesian index (tuple of ints)
        """
        return np.unravel_index(lin_idx, self.block_starts.shape[:-1])

    def cartesian_to_linear(self, ca_idx):
        """
        Convert block cartesian index in linear index.
        Example: in a 3D stack split in 2x2x3 blocks

        self.cartesian_to_linear0,0,0) = 0  # first block
        bs.cartesian_to_linear(1,1,2) = 11  # last block

        :param ca_idx: block cartesian index (tuple of ints)
        :return: block linear index (int)
        """
        return np.ravel_multi_index(ca_idx, self.block_starts.shape[:-1])

    def __getitem__(self, item):
        """
        :param item:
        :return:
        """
        # TODO make less brittle, support also indexing by tuples

        # TODO decide what should be returned: slices are tricky
        # with multiprocessing
        if isinstance(item, int):
            idx = self.linear_to_cartesian(item)
            return tuple(
                slice(s, e)
                for s, e in zip(self.block_starts[idx], self.block_ends[idx])
            )

    def neighbour_blocks(self, i_block, dims=None):
        """
        Return neighbouring blocks across given dimensions
        :param i_block:
        :param dims:
        :return:
        """
        block_idx = self.linear_to_cartesian(i_block)
        act_dims = np.ones(self.n_dims, dtype=np.bool)
        if dims is not None:
            act_dims[dims] = True

        neighbors = []
        for idx_neighbour in product(
            *[
                (
                    range(
                        max(block_idx[i_dim] - 1, 0),
                        min(block_idx[i_dim] + 1, self.block_starts.shape[i_dim]),
                    )
                    if act_dims[i_dim]
                    else [block_idx[i_dim]]
                )
                for i_dim in range(self.n_dims)
            ]
        ):
            if idx_neighbour != block_idx:
                neighbors.append(idx_neighbour)
        if neighbors:
            return np.ravel_multi_index(
                np.stack(neighbors, 1), self.block_starts.shape[:-1]
            )
        else:
            return np.array([])

    def blocks_to_take(self, start_take, end_take):
        """
        Find which blocks to take to cover the range:
        :param start_take: starting points in the N dims (tuple)
        :param end_take: ending points in the N dims (tuple)
        :return: tuple of tuples with the extremes of blocks to take in N dims;
                 starting index of data in the first block;
                 ending index of data in the last block.
        """
        n_dims = len(start_take)
        block_slices = []
        take_block_s_idx = []
        take_block_e_idx = []
        for i_dim, (start, end) in enumerate(zip(start_take, end_take)):
            axis_index = tuple(
                0 if i != i_dim else slice(None) for i in range(n_dims)
            ) + (i_dim,)
            s = max(
                0,
                min(
                    np.searchsorted(self.block_starts[axis_index], start) - 1,
                    len(self.block_starts[axis_index]) - 1,
                ),
            )
            e = np.searchsorted(self.block_starts[axis_index], end)
            block_start = start - self.block_starts[axis_index][s]
            block_end = end - self.block_starts[axis_index][e - 1]

            block_slices.append((s, e))
            take_block_s_idx.append(block_start)
            take_block_e_idx.append(block_end)
        return block_slices, take_block_s_idx, take_block_e_idx

    @staticmethod
    def block_to_slices(block):
        return tuple(slice(lb, rb) for lb, rb in block)

    def centres(self):
        return (self.block_ends + self.block_starts) / 2

    def block_containing_coords(self, coords):
        """
        Find the linear index of a block containing the given coordinates

        :param coords: a tuple of the coordinates
        :return:
        """
        dims = []
        for ic, c in enumerate(coords):
            # Create a tuple with the starting points on current dimension
            # for all the blocks:
            starts = self.block_starts[
                tuple(slice(None) if i == ic else 0 for i in range(self.n_dims)) + (ic,)
            ]

            # find in which position our guy should be ordered, correcting
            # for 0 value:
            dims.append(max((np.searchsorted(starts, c)) - 1, 0))
        return dims

    def drop_dim(self, dim_to_drop):
        """
        Return a new BlockSplitter object with a dimension dropped,
        useful for getting spatial from spatio-temporal blocks.

        :param dim_to_drop: dimension to be dropped (int)
        :return: new BlockSplitter object
        """
        drop_ith = lambda xs: tuple(x for i, x in enumerate(xs) if i != dim_to_drop)
        return Blocks(
            drop_ith(self.shape_full),
            shape_block=drop_ith(self.shape_block),
            padding=drop_ith(self.padding),
            crop=drop_ith(self.crop),
        )

    def serialize(self):
        """
        Returns a dictionary with a complete description of the
        BlockSplitter, e.g. to save its structure as json file.
        :return:
        """
        # TODO it should be possible to initialize the BlockSplitter from
        # this dictionary!
        return dict(
            shape_full=self.shape_full,
            shape_block=self.shape_block,
            crop_start=tuple(c[0] for c in self.crop),
            crop_end=tuple(c[1] for c in self.crop),
            padding=self.padding,
        )



#!/usr/bin/env python

"""Tests for `split_dataset` package."""

import unittest
import shutil
import numpy as np
import tempfile
from split_dataset import save_to_split_dataset


class TestSplitDataset(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_SplitDataset(self):
        dims = [(10, 3, 3, 3), (5, 5, 5), (5, 5), (1, 5, 5, 5)]
        block_sizes = [(2, None, None, None), (1, None, 3), (2, None), (None, 2, 5, 5)]
        all_slices = [
            [(slice(3, 8), slice(None))],
            [(slice(0, 1),), (slice(0, 2), slice(0, 1), slice(None))],
            [slice(0, 2)],
            [
                (slice(0, 1),),
                (slice(0, 2), slice(0, 1), slice(None)),
                (0, slice(0, 2), slice(0, 1)),
            ],
        ]

        for i, (di, bs, slices) in enumerate(zip(dims, block_sizes, all_slices)):
            test_data = np.arange(np.product(di)).reshape(di)

            sd = save_to_split_dataset(
                test_data,
                block_size=bs,
                root_name=self.test_dir,
                prefix="te{:02d}".format(i),
            )
            for sl in slices:
                a = sd[sl]
                b = test_data[sl]
                np.testing.assert_equal(
                    a,
                    b,
                    err_msg="Testing "
                    + str(di)
                    + " "
                    + str(sl)
                    + " of shape "
                    + str(a.shape)
                    + " and shape"
                    + str(b.shape),
                )

